#include "DBPool.hh"
#include "DBPooledConnection.hh"
#include "../config/DBHostConfig.hh"
#include "../postgres/DBPostgresPooledConnection.hh"
#include "../postgres/DBPostgresWriterConnection.hh"

/**
 * R-O connection, handled by DBPostgresPooledConnection.
 * This kind of pool has SEVERAL connections with ONE host.
 */

DBPool::DBPool(string daemon_user, string daemon_password, string daemon_auth_mode, int initDbConnections, DBHostConfig* host, string id) {

  // We save this data because maybe we'll need to make more connections.

  _daemon_user = daemon_user;
  _daemon_password = daemon_password;
  _daemon_auth_mode = daemon_auth_mode;
  _host = host;

  _used_conn_index=0;
  _tested_conn_index=0;

  _id = id;

  // We create the minimum number of connections
  // Firstly they will be Postgres connections, but
  // could be anyone that extends DBPooledConnection.

  cout << "    Postgres Pool in " << host->getName() << ":" << host->getDbPort() << "/" << host->getDbName() << " as "<< host->getDbUser() << ": ";
  for (int i=0; i< initDbConnections; i++) {
			
    char c[3];
    sprintf(c,"%d",i);
    c[2]=0;
    string id2(c);

 
    DBPostgresPooledConnection *post = new DBPostgresPooledConnection(_daemon_user,
								      _daemon_password,
								      _daemon_auth_mode,
								      _host->getDbUser(),
								      _host->getDbPassword(),
								      _host->getDbName(),
								      _host->getName(),
								      _host->getDbPort(),
								      id2);
     

    if (post->isValid()) {
   
      cout << " " << i;

    } else {
 
      cout << " X";

    }

    post->setId(id+"::"+c);
    _connections.push_back(post);
      
  }
  cout << endl;
  _nConns = _connections.size();

}


/**
 * R-W connection, handled by DBPostresWriterConnection.
 * Here we're creating a Pool that has ONE connection in EACH of the different hosts. The writes
 * are doing in parallel between the client and all the hosts.
 */

DBPool::DBPool(string daemon_user, string daemon_password, string daemon_auth_mode, vector<string> users, vector<string> passwords, vector<string>dbs, vector<string> hosts, vector<int>ports,  string id, bool writerTransactionPerRequest=false) {

  // Login accept data.
  _daemon_user = daemon_user;
  _daemon_password = daemon_password;

  // Load adapted conn numbers.
  //_minConn = 1;
  //_nconns = 1;
  //_maxConn = 1;

  // Conn index: Always 0, because this kind of pool only has one (though multiplexed) virtual connection.
  _used_conn_index=0;
  _tested_conn_index=0;

  _id = id;

  cout << "    Postgres Writing Pool in: ";

  for (vector<string>::iterator p = hosts.begin(); p!=hosts.end(); ++p) {
    cout << *p << " ";
  }
	      

  DBPostgresWriterConnection *post = new DBPostgresWriterConnection(daemon_user,
								    daemon_password,
								    daemon_auth_mode,
								    users,
								    passwords,
								    dbs,
								    hosts,
								    ports,
								    id,
								    writerTransactionPerRequest);


    if (post->isValid()) {
   
      cout << id << endl;

    } else {
 
      cout << "X" << endl;

    }

    post->setId(id);
    _connections.push_back(post);
    _nConns = _connections.size();
      
}


DBPool::~DBPool() {

  //cout << "Destroying DBPool()" << endl;

}

/** 
 * Here we've got to define the algorithm to choose a connection for test.
 * This method will be called by the connection reaper thread, which searches
 * for invalid or stuck connections.
 * There could be three cases:
 * a) !valid: urgent check and recover
 * b) valid && !free && (time_use > timeout): check and free.
 * c) valid && free: routine check.
 */

DBPooledConnection *DBPool::getConnectionForTest() {

  unsigned int connection;

  // Lock while computing next connection. (Just needed
  // if we've got several reapers).
  
  _tested_conn_index=((_tested_conn_index+1)>=_nConns)?0:(_tested_conn_index+1);

  connection = _tested_conn_index;
  // Unlock
  
  ACE_DEBUG((LM_DEBUG,"(%t %T) DBPool::getConnectionForTest(): Lets see if pool %s, Conn %d... \n ",_id.c_str(),connection));
  
  return _connections[connection]->testable();
     
	
}


// We should do one pass over every connection in the DBPool to
// get one that is in state "usable()==true".
// THIS METHOD FAILS IF WE'VE ONLY GOT *ONE* CONNECTION: CHECK WHY!!!

DBPooledConnection *DBPool::getPooledConnection(int strategy) {


  DBPooledConnection* db;

  switch (strategy) {

  case 0:
    
    for (unsigned int i=0; i<_nConns; i++) {

      // Lock while computing next connection.
      _used_conn_mutex.acquire();
      _used_conn_index=((_used_conn_index+1)>=_nConns)?0:(_used_conn_index+1);
      _used_conn_mutex.release();
      // Unlock

      ACE_DEBUG((LM_DEBUG,"(%t %T) DBPool::getPooledConnection(): Checking Pool %s, Conn %d, NConns %d. \n ",_id.c_str(),_used_conn_index, _nConns));
      db = _connections[_used_conn_index]->usable();
      if (db) return db;

    }
    
    return NULL;
     
	
  default:
    return NULL;

  }

}

// This method increases the number of connections (used only when READER DBPool)

int DBPool::grow(int growConns) {

  for (unsigned int i=_nConns; i< _nConns+growConns; i++) {

    char c[2];
    sprintf(c,"%d",i);

    DBPostgresPooledConnection *post = new DBPostgresPooledConnection(_daemon_user,
								      _daemon_password,
								      _daemon_auth_mode,
								      _host->getDbUser(),
								      _host->getDbPassword(),
								      _host->getDbName(),
								      _host->getName(),
								      _host->getDbPort(),
								      _id+"::"+c);

    if (post->isValid()) {
      cout << " " << i;
    } else {
      cout << " X";
    }

    _connections.push_back(post);
      
  }
  cout << endl;
  _nConns += growConns;
  return _nConns;


}

// This method decreases the number of connections (used only when READER DBPool)
// We have to find a way to make sure the conns are not being used when we delete them (dbc->usable())
// and to make sure the conn number remains stable.

int DBPool::shrink(int shrinkConns) {

  for(int i=0; i<abs(shrinkConns);) {

    DBPooledConnection* dbc = _connections.back();
    if (dbc->usable()) {
      _used_conn_mutex.acquire();
      _nConns -= 1;
      dbc->closeConnection();
      _connections.pop_back();
      delete dbc;
      _used_conn_mutex.release();
      i++;
    } else {
      ACE_DEBUG((LM_ERROR,"(%t,%T) DBPool::shrink Could not delete connection (still in use or invalid)\n"));
      ACE_OS::sleep(1);
    }

  }
  return _nConns;

}


